"""
Module to implement a custom connection class for Bluesky accounts,
supporting reading and writing statuses from the authenticated user
via their handle and password.
"""

import logging
import time
from datetime import datetime
from typing import Final, Optional, Union
from urllib.parse import urlparse

import requests
from atproto import Client
from atproto_client.exceptions import BadRequestError  # type: ignore
from atproto_client.models import (  # type: ignore
    AppBskyEmbedExternal,
    AppBskyEmbedImages,
    AppBskyEmbedRecord,
    AppBskyEmbedRecordWithMedia,
    AppBskyEmbedVideo,
    AppBskyRichtextFacet,
)
from atproto_client.models.common import XrpcError  # type: ignore
from bs4 import BeautifulSoup, Tag

from barkr.connections.base import Connection, ConnectionMode
from barkr.models import Message
from barkr.utils import (
    REQUESTS_EMBED_GET_TIMEOUT,
    REQUESTS_HEADERS,
    extract_urls_from_text,
)

logger = logging.getLogger()

BLUESKY_MAX_MESSAGE_LENGTH: Final[int] = 300


class BlueskyConnection(Connection):
    """
    Custom connection class for Bluesky accounts,
    supporting reading and writing statuses from the authenticated user.

    Requires handle and an app password for authentication.
    """

    def __init__(
        self, name: str, modes: list[ConnectionMode], handle: str, password: str
    ) -> None:
        """
        Initializes the connection with a name and a list of modes
        and sets up the initial connection between the client and Bluesky
        for the given user.

        NOTE: it is recommended to use an app password instead of the user's password.
        ref: https://bsky.app/settings/app-passwords

        :param name: The name of the connection
        :param modes: A list of modes for the connection
        :param handle: The handle of the authenticated user
        :param password: The app password of the authenticated user
        """

        super().__init__(name, modes)

        logger.info(
            "Initializing Bluesky (%s) connection for user %s",
            self.name,
            handle,
        )

        self.service = Client()
        self.service.login(handle, password)
        self.handle: str = handle

        logger.info(
            "Bluesky (%s) connection initialized! (User handle: %s)",
            self.name,
            self.handle,
        )

        user_feed = self.service.app.bsky.feed.get_author_feed({"actor": handle}).feed
        if user_feed:
            # Set the initial min_id to the most recent post's indexed_at,
            # which is a UTC timestamp string
            self.min_id: Optional[str] = user_feed[0].post.indexed_at
            logger.info("Bluesky (%s) initial min_id: %s", self.name, self.min_id)
        else:
            self.min_id = None
            logger.info("Bluesky (%s) initial min_id not set.", self.name)

    def _fetch(self) -> list[Message]:
        """
        Fetches messages from the authenticated user's account.

        :return: A list of messages
        """

        messages: list[Message] = []

        user_feed = self.service.app.bsky.feed.get_author_feed(
            {"actor": self.handle}
        ).feed
        if user_feed:
            for feed_view in user_feed:
                post = feed_view.post

                # Ignoring reposts
                if post.viewer is not None and post.viewer.repost is not None:
                    continue

                # Ignoring replies
                if post.record.reply is not None:
                    continue

                if self.min_id is None or datetime.fromisoformat(
                    post.indexed_at
                ) > datetime.fromisoformat(self.min_id):
                    record = post.record
                    if (embed := record.embed) is not None:
                        text = self._process_text_with_embed(record.text, embed)
                    else:
                        text = record.text

                    messages.append(Message(id=post.indexed_at, message=text))

        if messages:
            self.min_id = messages[0].id
            logger.info("Bluesky (%s) has %s new messages.", self.name, len(messages))
        else:
            logger.info("Bluesky (%s) has no new messages.", self.name)

        return messages

    def _post(self, messages: list[Message]) -> list[str]:
        """
        Posts the given messages to the authenticated user's account.

        :param messages: The messages to post
        :return: A list of IDs of the posted messages
        """

        posted_message_ids: list[str] = []

        for message in messages:
            if len(message.message) > BLUESKY_MAX_MESSAGE_LENGTH:
                logger.warning(
                    "Message length exceeds Bluesky (%s) "
                    "maximum length (%s), skipping: %s",
                    self.name,
                    BLUESKY_MAX_MESSAGE_LENGTH,
                    message.message,
                )
                continue

            embed, facets = self._generate_post_embed_and_facets(message.message)
            try:
                created_record = self.service.send_post(
                    text=message.message, embed=embed, facets=facets if facets else None
                )
            except BadRequestError as e:
                # We could be trying to create an embed that is too large,
                # let's recover
                error_response = e.response
                content = error_response.content

                if isinstance(content, XrpcError) and content.error == "BlobTooLarge":
                    logger.warning(
                        "Bluesky (%s) post failed due to embed size, "
                        "reattempting post without embed.",
                        self.name,
                    )
                    created_record = self.service.send_post(
                        text=message.message,
                        embed=None,
                        facets=facets if facets else None,
                    )
                else:
                    logger.error(
                        "Bluesky (%s) post failed with error: %s",
                        self.name,
                        content,
                    )
                    raise e

            created_uri = created_record.uri

            # NOTE: introducing an artificial delay to ensure the post is indexed
            # before fetching the post details
            time.sleep(1)

            post_details = self.service.get_posts([created_uri]).posts[0]
            indexed_at = post_details.indexed_at

            logger.info(
                "Posted message %s to Bluesky (%s) connection (URI: %s, Indexed At: %s)",
                message.message,
                self.name,
                created_uri,
                indexed_at,
            )

            self.min_id = indexed_at
            posted_message_ids.append(indexed_at)

        return posted_message_ids

    def _generate_post_embed_and_facets(
        self, text: str
    ) -> tuple[Optional[AppBskyEmbedExternal.Main], list[AppBskyRichtextFacet.Main]]:
        """
        If a link is present on the text for a post to be created,
        creates an Embed object for the link.

        This allows Bluesky to display the link as a preview.

        Also generate facets for the link to be used in the post.

        :param text: The text of the post
        :return: A tuple containing the Embed object and a list of facets
        """

        embed = None
        facets: list[AppBskyRichtextFacet.Main] = []

        for url in extract_urls_from_text(text):
            # Hit the URL to check if it is valid and extract title and description
            try:
                # Using a very short timeout, we don't want to spend too much time here
                response = requests.get(
                    url, timeout=REQUESTS_EMBED_GET_TIMEOUT, headers=REQUESTS_HEADERS
                )
            except requests.RequestException:
                continue

            if response.status_code == 200:
                if embed is None:
                    # Parse the response to extract title, description and image
                    soup = BeautifulSoup(response.content, "html.parser")
                    title = soup.title.string if soup.title else urlparse(url).netloc
                    description = url
                    thumbnail_blob = None

                    # Extract the description from the meta tag
                    if (
                        meta_description := _get_meta_tag_from_html_metadata(
                            soup, "og:description"
                        )
                    ) is not None:
                        description = meta_description

                    # Extract the image from the meta tag
                    if (
                        image := _get_meta_tag_from_html_metadata(soup, "og:image")
                    ) is not None:
                        # Fetching the image and reuploading to Bluesky
                        try:
                            img_data = requests.get(
                                image,
                                timeout=REQUESTS_EMBED_GET_TIMEOUT,
                                headers=REQUESTS_HEADERS,
                            ).content
                        except requests.RequestException as e:
                            logger.warning(
                                "Failed to fetch image from %s: %s", image, e
                            )
                        else:
                            thumbnail_blob = self.service.upload_blob(img_data).blob

                    # Prepare the Embed object
                    embed = AppBskyEmbedExternal.Main(
                        external=AppBskyEmbedExternal.External(
                            uri=url,
                            title=title,
                            description=description,
                            thumb=thumbnail_blob,
                        )
                    )

                # Create a facet for the link
                url_index = text.index(url)
                facets.append(
                    AppBskyRichtextFacet.Main(
                        features=[
                            AppBskyRichtextFacet.Link(
                                uri=url,
                            )
                        ],
                        index=AppBskyRichtextFacet.ByteSlice(
                            byte_start=len(text[:url_index].encode("utf-8")),
                            byte_end=len(text[: url_index + len(url)].encode("utf-8")),
                        ),
                    )
                )

        return embed, facets

    def _process_text_with_embed(
        self,
        text: str,
        embed: Optional[
            Union[
                AppBskyEmbedExternal.Main,
                AppBskyEmbedRecord.Main,
                AppBskyEmbedImages.Main,
                AppBskyEmbedVideo.Main,
                AppBskyEmbedRecordWithMedia.Main,
            ]
        ],
    ) -> str:
        """
        Handles the special case where a Bluesky post contains a link to an embedded
        resources that is not fully rendered as part of the text.

        Leveraging the Embed object, reconstructs the text to include
        the full URL to the resource.

        For example, when posting the URL
        https://open.spotify.com/track/0ElVpg9XIswx3XWs6kUj6a?si=0015d86587524ef9
        the text is trimmed to open.spotify.com/track/0ElVpg... but the
        Embed object contains the full URL.

        :param text: The original text of the post
        :param embed: The Embed object containing the link
        :return: The reconstructed text with the full URL
        """

        if embed is None:
            return text

        # Depending on the type of embed, we get the URL
        # from the corresponding field
        if isinstance(embed, AppBskyEmbedExternal.Main):
            url = embed.external.uri
        else:
            return text

        # We now want to find the word in the text that is contained
        # in the URL, and we only care for the _longest_ word
        # if there are multiple matches
        matching_word = ""
        for word in text.split():
            if word.replace("...", "") in url:
                if len(word) > len(matching_word):
                    matching_word = word

        if not matching_word:
            return text

        return text.replace(matching_word, url)


def _get_meta_tag_from_html_metadata(
    soup: BeautifulSoup, tag_name: str
) -> Optional[str]:
    """
    Extracts the content of meta tag from the HTML metadata of a page.

    If there are multiple meta tags with the same property,
    only the first one is returned.

    :param soup: The BeautifulSoup object containing the HTML metadata
    :param tag_name: The name of the meta tag to extract
    :return: The meta tag content if found, otherwise None
    """

    tag = soup.find("meta", attrs={"property": tag_name})
    if isinstance(tag, Tag):
        tag_content = tag["content"]

        if isinstance(tag_content, list):
            tag_content = tag_content[0]

        return tag_content

    return None



"""
Utility functions for the Barkr package.
"""

import re
from time import sleep
from typing import Any, Callable, Final

REQUESTS_EMBED_GET_TIMEOUT: Final[int] = 3

REQUESTS_HEADERS: Final[dict[str, str]] = {
    "User-Agent": (
        "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:20.0) " "Gecko/20100101 Firefox/20.0"
    )
}


def wrap_while_true(
    callback: Callable[[], Any], sleep_interval: int
) -> Callable[[], Any]:
    """
    Wrap a function to be called within a `while True` scope,
    sleeping for a given interval between calls.

    :param callback: The function to be called within the loop
    :param sleep_interval: The interval to sleep between calls, in seconds
    :return: The wrapped function
    """

    def wrapper() -> None:
        while True:
            callback()
            sleep(sleep_interval)

    return wrapper


def extract_urls_from_text(text: str) -> list[str]:
    """
    Given a text string, extract and return all URLs
    found within it in a list.

    URLs are not validated as reachable and no connection
    is made to them, extraction is done using regex.

    :param text: The text to extract URLs from
    :return: A list of extracted URLs
    """

    return re.findall(r"http[s]?://[^\s]+", text)



"""
Module that implements the main loop of the Barkr application,
enabling users to instance the Barkr class with their own connections
to set crossposting among multiple channels.
"""

import logging
from threading import Lock, Thread
from typing import Optional

from barkr.connections import Connection, ConnectionMode
from barkr.models import Message
from barkr.utils import wrap_while_true

logger = logging.getLogger()


class Barkr:
    """
    Wrapper for the main loop of the application.
    """

    polling_interval: int
    write_rate_limit: Optional[int]
    connections: list[Connection]
    message_queues: dict[str, list[Message]]
    message_queues_lock: Lock

    def __init__(
        self,
        connections: list[Connection],
        polling_interval: int = 10,
        write_rate_limit: Optional[int] = None,
    ) -> None:
        """
        Instantiate a Barkr object with a list of connections, as well as
        internal queues and locks.

        If a `write_rate_limit` is provided, Barkr will only write up to that
        many messages per write-thread polling interval.

        :param connections: A list of connections to be used by the Barkr instance
        :param polling_interval: The interval to wait between polling requests, in seconds
        :param write_rate_limit: (optional) The rate limit for writing messages
        """

        if not connections:
            raise ValueError("Must provide at least one connection!")

        if polling_interval < 1:
            raise ValueError("Polling interval must be at least 1 second!")

        if write_rate_limit is not None and write_rate_limit < 1:
            raise ValueError(
                "If specified, write rate limit must be at least 1 message!"
            )

        self.polling_interval: int = polling_interval
        self.write_rate_limit: Optional[int] = write_rate_limit

        logger.info(
            "Initializing Barkr instance with %s connection(s)...", len(connections)
        )
        self.connections = connections
        self.message_queues = {connection.name: [] for connection in connections}
        self.message_queues_lock = Lock()
        logger.info("Barkr instance initialized!")

    def read(self) -> None:
        """
        Read messages from all connections and add them to the message queues
        of other connections
        """

        for connection in self.connections:
            # Reading is only allowed for connections with the READ mode
            if ConnectionMode.READ not in connection.modes:
                continue

            messages = connection.read()

            if messages:
                with self.message_queues_lock:
                    for name in self.message_queues:
                        if name != connection.name:
                            self.message_queues[name] += messages
                            logger.info(
                                "Added %s message(s) from %s to %s queue",
                                len(messages),
                                connection.name,
                                name,
                            )

    def write(self) -> None:
        """
        Write messages from the message queues to all connections
        """

        for connection in self.connections:
            max_amount: int = self.write_rate_limit or len(
                self.message_queues[connection.name]
            )

            # Writing is only allowed for connections with the WRITE mode
            if ConnectionMode.WRITE in connection.modes:
                with self.message_queues_lock:
                    messages = self.message_queues[connection.name][:max_amount]
                    n_messages = len(messages)
                    logger.info(
                        "Writing %s message(s) from %s's queue "
                        "(%s messages remaining after that)",
                        n_messages,
                        connection.name,
                        len(self.message_queues[connection.name]) - n_messages,
                    )

                    if messages:
                        connection.write(messages)
                        logger.info(
                            "Posted %s message(s) from %s's queue",
                            n_messages,
                            connection.name,
                        )

            # Clear sent messages from the queue for the current connection
            with self.message_queues_lock:
                self.message_queues[connection.name] = self.message_queues[
                    connection.name
                ][max_amount:]

    def write_message(self, message: Message) -> None:
        """
        Write a single message to all connections. Barkr does not need
        to be running on a loop for this to work.


        Useful if your use case does not require a read-write loop,
        and for testing purposes.

        :param message: The message to be sent
        """

        logger.info("Writing message to all connections...")

        for connection in self.connections:
            if ConnectionMode.WRITE in connection.modes:
                connection.write([message])
                logger.info("Posted message to %s", connection.name)

        logger.info("Message posted to all connections!")

    def start(self) -> None:
        """
        Start the Barkr instance
        """

        logger.info("Starting Barkr!")

        read_thread = Thread(target=wrap_while_true(self.read, self.polling_interval))
        write_thread = Thread(target=wrap_while_true(self.write, self.polling_interval))

        read_thread.start()
        write_thread.start()
        logger.info("Barkr started!")

        read_thread.join()
        write_thread.join()
        logger.info("Barkr exiting!")



"""
Module to implement a generic, wrapper model for messages.

A message is a generic object that can be posted to a social media network
OR retrieved from a social media network, with an ID and a message body, as
well as (potentially) other metadata that each connection can decide
whether or not to take into account.
"""

from dataclasses import dataclass


@dataclass(frozen=True)
class Message:
    """
    A generic message object with an ID and a message body,
    as well as (potentially) other metadata that each connection
    can decide whether or not to take into account.
    """

    id: str
    message: str

    def has_content(self) -> bool:
        """
        Check if the message has content and therefore can be published.

        At this point, the only check is whether the message is empty or not.
        In the future, as Messages grow in complexity, this method may be
        extended to check for other conditions (e.g. accounting for
        a message with no text but with an image).

        :return: True if the message has content, False otherwise
        """

        return bool(self.message)



"""
Re-exporting all connection classes and enums from the connections submodules
for ease of use.

You can refer to each submodule for more information about the classes and enums,
and how to use them.
"""

from barkr.connections.base import Connection, ConnectionMode
from barkr.connections.bluesky import BlueskyConnection
from barkr.connections.discord import DiscordConnection
from barkr.connections.mastodon import MastodonConnection
from barkr.connections.mastodon_activity_bot import MastodonActivityBotConnection
from barkr.connections.rss import RSSConnection
from barkr.connections.telegram import TelegramConnection
from barkr.connections.twitter import TwitterConnection

__all__ = [
    "Connection",
    "TwitterConnection",
    "MastodonActivityBotConnection",
    "MastodonConnection",
    "ConnectionMode",
    "DiscordConnection",
    "BlueskyConnection",
    "TelegramConnection",
    "RSSConnection",
]



"""
Module to implement a custom connection class for ActivityBot,
an ActivityPub bot for Mastodon.

Supports writing statuses to Mastodon from the bot via
their API url and password.

ref: https://gitlab.com/edent/activity-bot
"""

import logging

import requests

from barkr.connections.base import Connection, ConnectionMode
from barkr.models import Message
from barkr.utils import REQUESTS_EMBED_GET_TIMEOUT, REQUESTS_HEADERS

logger = logging.getLogger()


class MastodonActivityBotConnection(Connection):
    """
    Custom connection class for ActivityBot, an ActivityPub bot for Mastodon.

    Supports writing statuses to Mastodon from the bot via
    their API url and password.
    """

    password: str
    api_url: str

    def __init__(
        self, name: str, modes: list[ConnectionMode], api_url: str, password: str
    ) -> None:
        """
        Initializes the connection with a name, API URL, and password.

        :param name: The name of the connection
        :param modes: A list of modes for the connection
        :param api_url: The API URL for the ActivityBot send action
        :param password: The password for the ActivityBot
        """
        super().__init__(name, modes)

        logger.info("Initializing MastodonActivityBot (%s) connection", self.name)
        if self.modes != [ConnectionMode.WRITE]:
            raise NotImplementedError(
                "MastodonActivityBotConnection only supports write mode."
            )

        self.api_url = api_url
        self.password = password

        logger.info(
            "MastodonActivityBot (%s) connection initialized! (API URL: %s)",
            self.name,
            self.api_url,
        )

    def _post(self, messages: list[Message]) -> list[str]:
        """
        Posts the given messages to ActivityPub via
        the ActivityBot send action.

        :param messages: The list of messages to post
        :return: A list of message IDs (empty for ActivityBot)
        """

        for message in messages:
            logger.info(
                "Posting message to ActivityBot (%s): %s", self.name, message.message
            )
            response = requests.post(
                self.api_url,
                data={"password": self.password, "content": message.message},
                headers=REQUESTS_HEADERS,
                timeout=REQUESTS_EMBED_GET_TIMEOUT,
            )

            if response.status_code != 200:
                logger.error(
                    "Failed to post message to ActivityBot (%s): %s",
                    self.name,
                    response.text,
                )

        return []
